/*
Copyright 2021 Plume Qi <plumeqi@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:
  
The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
	
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
#include "Splasher.h"
#pragma comment(lib, "Winmm.lib")

typedef struct tagSplashStruct
{
	HINSTANCE hInstance;        /*The instance handle to load from */
	LPCTSTR   lpszName;         /*The name of bitmap resource or file*/
	BOOL      bRes;             /*Use bitmap resource or not*/
	DWORD     dwWindowStyle;    /*Splash window style*/
	DWORD     dwARGB;           /*The flags passed to SetLayeredWindowAttributes. It will split into two: key color and alpha*/
	HBITMAP   hBitmap;          /*The image we are displaying*/
	LONG      bmpWidth;         /*Width of the image*/
	LONG      bmpHeight;        /*Height of the image*/
	HDC       hDCBitmap;        /*bitmap dc, double buffer*/
	LPTSTR    pszSound;         /*Splash sound file or resource*/
	BOOL      bResSound;        /*TRUE, splash sound is from resource; FALSE; splash sound is from file*/
	HWND      hwndOwner;        /*Our hidden window parent (causes this window not to have a entry in the task bar)*/
	HWND      hwndSplash;       /*Splash window handle*/
	RECT      rcText;           /*The rect for showing text*/
	TCHAR     szText[128];      /*The text will be displayed*/
	COLORREF  clrTextColor;     /*Text color*/
	int       nTextCount;       /*Text count*/
	HANDLE    hSplashThread;    /*Splash thread handle*/
	HANDLE    hSplashEvent;     /*Splash event handle*/
	BOOL      bThreadInitOK;    /*Splash window prepare OK*/
	BOOL      bOKToClose;       /*Ok to respond to WM_CLOSE messages*/
	DWORD     dwShowTime;       /*Splash windows will be destroyed after dwShowTime expired if loading time is less than dwShowTime*/
	DWORD     dwTheadStartTick; /*Tick count when splash thread start*/
	
}SPLASHSTRUCT, *PSPLASHSTRUCT;

static PSPLASHSTRUCT m_pSplashStruct = NULL; /*Splash data*/


static DWORD WINAPI SplashThread(LPVOID lpData);

static BOOL LoadSplashBitmap(HINSTANCE hInstance, LPCTSTR lpszName, BOOL bRes);
static void FreeSplashResource();

typedef BOOL (WINAPI *pSetLayeredWindowAttributes)(HWND, COLORREF, BYTE, DWORD);
static BOOL InitInstance();
static BOOL ExitInstance();
static BOOL CreateSplashWindow();
static BOOL CenterWindow(HWND hwndWindow);

typedef struct tagDecodeUINT
{
	UINT code;
	LRESULT (*Fxn)(HWND, UINT, WPARAM, LPARAM);
}DECODEUINT, *PDECODEUINT;

static LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
static LRESULT DoPaint(HWND, UINT, WPARAM, LPARAM);
static LRESULT DoCreate(HWND, UINT, WPARAM, LPARAM);
static LRESULT DoClose(HWND, UINT, WPARAM, LPARAM);
static LRESULT DoDestroy(HWND, UINT, WPARAM, LPARAM);
static LRESULT DoNcHitTest(HWND, UINT, WPARAM, LPARAM);
static LRESULT DoShowText(HWND, UINT, WPARAM, LPARAM);

const DECODEUINT arrMsg[] = 
{
		WM_PAINT,            DoPaint,
		WM_CREATE,           DoCreate,
		WM_CLOSE,            DoClose,
		WM_DESTROY,          DoDestroy,
		WM_NCHITTEST,        DoNcHitTest,
		WM_USER+101,         DoShowText,
};

static LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int i;
	/* Search message list to see if need to handle this message. If in list, call procedure. */
	for( i=0; i<(sizeof(arrMsg)/sizeof(DECODEUINT)); i++ )
	{
		if( message == arrMsg[i].code )
			return (*arrMsg[i].Fxn)(hWnd, message, wParam, lParam);
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

#ifndef CS_DROPSHADOW
#define CS_DROPSHADOW 0x00020000
#endif
#ifndef WS_EX_LAYERED
#define WS_EX_LAYERED 0x00080000
#endif

static BOOL CreateSplashWindow()
{
	WNDCLASSEX wcex;

	if( m_pSplashStruct == NULL )
		return FALSE;

	wcex.cbSize = sizeof(WNDCLASSEX); 

	wcex.style			= 0;
	wcex.lpfnWndProc	= DefWindowProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= m_pSplashStruct->hInstance;
	wcex.hIcon			= NULL;
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= NULL;
	wcex.lpszClassName	= _T("SplashOwnerWnd");
	wcex.hIconSm		= NULL;
	RegisterClassEx(&wcex);

	m_pSplashStruct->hwndOwner = CreateWindowEx(
		0,
		wcex.lpszClassName, 
		_T(""), 
		WS_POPUP,
		CW_USEDEFAULT, 
		CW_USEDEFAULT, 
		CW_USEDEFAULT, 
		CW_USEDEFAULT, 
		NULL, 
		NULL, 
		m_pSplashStruct->hInstance, 
		NULL);
	
	wcex.lpfnWndProc	= (WNDPROC)WndProc;
	wcex.lpszClassName	= _T("SplashWnd");
	if (m_pSplashStruct->dwWindowStyle & FLAG_DROPSHADOW )
		wcex.style = CS_DROPSHADOW;
	RegisterClassEx(&wcex);
	m_pSplashStruct->hwndSplash = CreateWindowEx(
		(m_pSplashStruct->dwWindowStyle & (FLAG_LWA_COLORKEY | FLAG_LWA_ALPHA))?(WS_EX_TOPMOST|WS_EX_LAYERED):WS_EX_TOPMOST,
		wcex.lpszClassName, 
		_T(""), 
		WS_POPUP | WS_VISIBLE,
		0, 
		0, 
		m_pSplashStruct->bmpWidth, 
		m_pSplashStruct->bmpHeight, 
		m_pSplashStruct->hwndOwner, 
		NULL, 
		m_pSplashStruct->hInstance, 
		NULL);

	if( m_pSplashStruct->hwndSplash == NULL )
		return FALSE;

	/*Make the window transparent if required*/
	if (m_pSplashStruct->dwWindowStyle & (FLAG_LWA_COLORKEY | FLAG_LWA_ALPHA) )
	{
		/*Make the Window transparent using the specified color*/

		HMODULE hModuleUser32 = NULL;
		pSetLayeredWindowAttributes Func_SetLayeredWindowAttributes = NULL;
		hModuleUser32 = LoadLibrary(_T("User32.dll"));
		Func_SetLayeredWindowAttributes = (pSetLayeredWindowAttributes)GetProcAddress(hModuleUser32, "SetLayeredWindowAttributes");
		
		Func_SetLayeredWindowAttributes(m_pSplashStruct->hwndSplash,
			m_pSplashStruct->dwARGB & 0x00FFFFFF, 
			(BYTE)((m_pSplashStruct->dwARGB & 0xFF000000)>>24), 
			(m_pSplashStruct->dwWindowStyle & (FLAG_LWA_COLORKEY | FLAG_LWA_ALPHA)) );

		if( hModuleUser32 )
			FreeLibrary(hModuleUser32);
	}

	return TRUE;
}

static void FreeSplashResource()
{
	if( m_pSplashStruct )
	{
		if( m_pSplashStruct->hwndSplash )
			DestroyWindow(m_pSplashStruct->hwndOwner);
		if( m_pSplashStruct->hwndSplash )
			DestroyWindow(m_pSplashStruct->hwndSplash);
		if( m_pSplashStruct->hBitmap )
			DeleteObject(m_pSplashStruct->hBitmap);
		if( m_pSplashStruct->hDCBitmap )
			DeleteDC(m_pSplashStruct->hDCBitmap);
		if( m_pSplashStruct->pszSound )
			LocalFree(m_pSplashStruct->pszSound);
		LocalFree((HLOCAL)m_pSplashStruct);
		m_pSplashStruct = NULL;
	}
}

static LRESULT DoPaint(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	HDC hDCPaint = NULL;
	PAINTSTRUCT Paint;

	hDCPaint = BeginPaint(hWnd, &Paint);

	if( m_pSplashStruct )
	{
		BitBlt(hDCPaint,0,0,m_pSplashStruct->bmpWidth,m_pSplashStruct->bmpHeight,m_pSplashStruct->hDCBitmap,0,0,SRCCOPY);
		SetBkMode(hDCPaint,TRANSPARENT);
		SetTextColor(hDCPaint,m_pSplashStruct->clrTextColor);
		DrawText(hDCPaint,m_pSplashStruct->szText,m_pSplashStruct->nTextCount,&(m_pSplashStruct->rcText),DT_LEFT);
	}

	EndPaint(hWnd, &Paint);
	return 1;
}

static LRESULT DoCreate(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	CenterWindow(hWnd);
	if( m_pSplashStruct && m_pSplashStruct->pszSound )
		PlaySound(m_pSplashStruct->pszSound, NULL, SND_ASYNC | (m_pSplashStruct->bResSound?SND_RESOURCE:SND_FILENAME));
	return 1;
}

static LRESULT DoClose(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if (m_pSplashStruct && m_pSplashStruct->bOKToClose)
	{
		/*Just request the parent window to close which will cause this window to be destroyed*/
		//PostMessage(m_wndOwner,WM_CLOSE,0,0);
		DestroyWindow(hWnd);
	}
	return 0;
}

static LRESULT DoDestroy(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if( m_pSplashStruct )
	{
		DestroyWindow(m_pSplashStruct->hwndOwner);
		m_pSplashStruct->hwndOwner = NULL;
		m_pSplashStruct->hwndSplash = NULL;
	}
	PostQuitMessage(0);
	return 0;
}

static LRESULT DoNcHitTest(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if (m_pSplashStruct && (m_pSplashStruct->dwWindowStyle & FLAG_DROPSHADOW) )
		return HTCAPTION;
	
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

static LRESULT DoShowText(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if (m_pSplashStruct )
	{
		InvalidateRect(hWnd,&(m_pSplashStruct->rcText),FALSE);
	}

	return 0;
}

static BOOL LoadSplashBitmap(HINSTANCE hInstance, LPCTSTR lpszName, BOOL bRes)
{
	BITMAP bmpInfo;
	memset(&bmpInfo,0,sizeof(BITMAP));

	if( m_pSplashStruct == NULL )
		return FALSE;

	if( bRes )
	{
		m_pSplashStruct->hBitmap = (HBITMAP)LoadImage(hInstance, 
			lpszName,
			IMAGE_BITMAP, 
			0, 
			0, 
			LR_DEFAULTSIZE | LR_DEFAULTCOLOR); 	 
	}
	else
	{
		m_pSplashStruct->hBitmap = (HBITMAP)LoadImage(NULL, 
			lpszName,
			IMAGE_BITMAP, 
			0, 
			0, 
			LR_DEFAULTSIZE | LR_DEFAULTCOLOR|LR_LOADFROMFILE );
	}

	if( m_pSplashStruct->hBitmap == NULL )
		return FALSE;
    
	GetObject(m_pSplashStruct->hBitmap, sizeof(BITMAP), &bmpInfo);
	m_pSplashStruct->bmpWidth = bmpInfo.bmWidth;
	m_pSplashStruct->bmpHeight = bmpInfo.bmHeight;

	m_pSplashStruct->hDCBitmap = CreateCompatibleDC(NULL);

	if( m_pSplashStruct->hDCBitmap == NULL )
		return FALSE;

	SelectObject(m_pSplashStruct->hDCBitmap,m_pSplashStruct->hBitmap);

	/*Calculate text rect*/
	m_pSplashStruct->rcText.left = 0;
	m_pSplashStruct->rcText.top = 0;
	DrawText(m_pSplashStruct->hDCBitmap, _T("X"), 1, &(m_pSplashStruct->rcText), DT_CALCRECT);
	m_pSplashStruct->rcText.left = 16;
	m_pSplashStruct->rcText.top = bmpInfo.bmHeight - m_pSplashStruct->rcText.bottom-16;
	m_pSplashStruct->rcText.right = bmpInfo.bmWidth - 16;
	m_pSplashStruct->rcText.bottom = bmpInfo.bmHeight - 16;

	return TRUE;
}

BOOL CenterWindow(HWND hwndWindow)
{
	HWND hwndParent;
	RECT rectWindow, rectParent;

	// make the window relative to its parent
	if ((hwndParent = GetParent(hwndWindow)) != NULL)
	{
		int nWidth = 0;
		int nHeight = 0;
		int nX = 0;
		int nY = 0;
		int nScreenWidth = 0;
		int nScreenHeight = 0;

		GetWindowRect(hwndWindow, &rectWindow);
		GetWindowRect(hwndParent, &rectParent);
		if( (rectParent.right-rectParent.left) == 0 ||
			(rectParent.bottom-rectParent.top) == 0 )
		{
			rectParent.left = 0;
			rectParent.right = GetSystemMetrics(SM_CXSCREEN);
			rectParent.top = 0;
			rectParent.bottom = GetSystemMetrics(SM_CYSCREEN);
		}

		nWidth = rectWindow.right - rectWindow.left;
		nHeight = rectWindow.bottom - rectWindow.top;

		nX = ((rectParent.right - rectParent.left) - nWidth) / 2 + rectParent.left;
		nY = ((rectParent.bottom - rectParent.top) - nHeight) / 2 + rectParent.top;

		nScreenWidth = GetSystemMetrics(SM_CXSCREEN);
		nScreenHeight = GetSystemMetrics(SM_CYSCREEN);

		// make sure that the dialog box never moves outside of the screen
		if (nX < 0) nX = 0;
		if (nY < 0) nY = 0;
		if (nX + nWidth > nScreenWidth) nX = nScreenWidth - nWidth;
		if (nY + nHeight > nScreenHeight) nY = nScreenHeight - nHeight;

		MoveWindow(hwndWindow, nX, nY, nWidth, nHeight, FALSE);

		return TRUE;
	}

	return FALSE;
}

static BOOL InitInstance()
{
	if( m_pSplashStruct == NULL )
		return FALSE;

	/*Create Splash window*/
	CreateSplashWindow();

	/*Signal the event which indicates that this InitInstance function has run*/
	m_pSplashStruct->bThreadInitOK = TRUE;
	SetEvent(m_pSplashStruct->hSplashEvent);

	return TRUE;
}

static BOOL ExitInstance()
{
	if( m_pSplashStruct )
	{
		if( m_pSplashStruct->hwndSplash )
		{
			DestroyWindow(m_pSplashStruct->hwndOwner);
			m_pSplashStruct->hwndOwner = NULL;
		}
		if( m_pSplashStruct->hwndSplash )
		{
			DestroyWindow(m_pSplashStruct->hwndSplash);
			m_pSplashStruct->hwndSplash = NULL;
		}
	}

	return TRUE;
}

static DWORD WINAPI SplashThread(LPVOID lpData)
{
	MSG msg;
	
	if( m_pSplashStruct == NULL )
		return 0;
	InitInstance();
	
	// Main message loop:
	while (GetMessage(&msg, NULL, 0, 0)) 
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return (int) msg.wParam;
}

BOOL Splash_Create(HINSTANCE hInstance, LPCTSTR lpszName, BOOL bRes, DWORD dwFlag, DWORD dwARGB,LPCTSTR lpszSound,BOOL bResSound,DWORD dwShowTime)
{
	m_pSplashStruct = (PSPLASHSTRUCT)LocalAlloc(LPTR,sizeof(SPLASHSTRUCT));
	if( m_pSplashStruct == NULL )
		return FALSE;

	m_pSplashStruct->hInstance = hInstance;
	m_pSplashStruct->lpszName = lpszName;
	m_pSplashStruct->bRes = bRes;
	m_pSplashStruct->dwWindowStyle = dwFlag;
	m_pSplashStruct->dwARGB = dwARGB;
	m_pSplashStruct->dwShowTime = dwShowTime;
	if( lpszSound )
	{
		m_pSplashStruct->pszSound = (LPTSTR)LocalAlloc(LPTR, (_tcslen(lpszSound)+1)*sizeof(TCHAR));
		_tcscpy(m_pSplashStruct->pszSound,lpszSound);
	}
	m_pSplashStruct->bResSound = bResSound;

	if( !LoadSplashBitmap(hInstance,lpszName,bRes) )
	{
		FreeSplashResource();
		return FALSE;
	}

	m_pSplashStruct->hSplashEvent = CreateEvent(NULL,FALSE,FALSE,NULL);
	m_pSplashStruct->dwTheadStartTick = GetTickCount();
	m_pSplashStruct->hSplashThread = CreateThread(NULL, 0, SplashThread, NULL, CREATE_SUSPENDED, NULL);
	if( m_pSplashStruct->hSplashThread == NULL )
	{
		FreeSplashResource();
		return FALSE;
	}

	ResumeThread(m_pSplashStruct->hSplashThread);
	WaitForSingleObject(m_pSplashStruct->hSplashEvent, INFINITE);

	if( !(m_pSplashStruct->bThreadInitOK) )
	{
		FreeSplashResource();
		return FALSE;
	}

	return TRUE;
}

void Splash_ShowText(LPCTSTR lpszText,COLORREF clrTextColor)
{
	if( m_pSplashStruct && m_pSplashStruct->hwndSplash )
	{
		if( lpszText == NULL )
		{
			m_pSplashStruct->szText[0] = _T('\0');
			m_pSplashStruct->nTextCount = 0;
		}
		else
		{
			size_t len = _tcslen(lpszText);
			m_pSplashStruct->nTextCount = (int)(len>=128?127:len);
			_tcsncpy(m_pSplashStruct->szText,lpszText,m_pSplashStruct->nTextCount);
		}
		m_pSplashStruct->clrTextColor = clrTextColor;
		PostMessage(m_pSplashStruct->hwndSplash, WM_USER+101,0,0);
	}
}

BOOL Splash_Close(HWND hWndActive)
{
	/*What will be the return value from this function (assume the worst)*/
	BOOL bSuccess = FALSE;
	if( m_pSplashStruct == NULL )
		return FALSE;

	if( m_pSplashStruct->hSplashThread != NULL )
	{
		HWND hWnd = NULL;

		/*check expired time*/
		while( (GetTickCount()-m_pSplashStruct->dwTheadStartTick) < m_pSplashStruct->dwShowTime )
			Sleep(100);

		/*Set the focus back to the correct window*/
		if( hWndActive )
			hWnd = hWndActive;
		else
			hWnd = GetActiveWindow();
		SetFocus(hWnd);
		SetForegroundWindow(hWnd);
		
		if( m_pSplashStruct->hwndSplash != NULL )
		{
			m_pSplashStruct->bOKToClose = TRUE;
			PostMessage(m_pSplashStruct->hwndSplash,WM_CLOSE,0,0);
		}
		WaitForSingleObject(m_pSplashStruct->hSplashThread,INFINITE);
		FreeSplashResource();
		m_pSplashStruct = NULL;
		bSuccess = TRUE;
	}

	return bSuccess;
}